@@ -2,7 +2,7 @@
  * @file    editor.c
  * @brief   
  *
- * Copyright (C) 2010 Gummi-Dev Team <alexvandermey@gmail.com>
+ * Copyright (C) 2010-2011 Gummi-Dev Team <alexvandermey@gmail.com>
  * All Rights reserved.
  * 
  * Permission is hereby granted, free of charge, to any person
@@ -45,9 +45,15 @@
 #include <gtk/gtk.h>
 
 #include "configfile.h"
+#include "constants.h"
 #include "environment.h"
 #include "utils.h"
 
+static void on_inserted_text(GtkTextBuffer *textbuffer,GtkTextIter *location,
+                             gchar *text,gint len, gpointer user_data);
+static void on_delete_range(GtkTextBuffer *textbuffer,GtkTextIter *start,
+                             GtkTextIter *end, gpointer user_data);
+                             
 const gchar style[][3][20] = {
     { "tool_bold", "\\textbf{", "}" },
     { "tool_italic", "\\emph{", "}" },
@@ -57,24 +63,24 @@ const gchar style[][3][20] = {
     { "tool_right", "\\begin{flushright}", "\\end{flushright}"}
 };
 
-GuEditor* editor_init (GuMotion* mc) {
+GuEditor* editor_new (GuMotion* mc) {
     GuEditor* ec = g_new0 (GuEditor, 1);
 
     /* File related member initialization */
     ec->workfd = -1;
     ec->fdname = NULL;
     ec->filename = NULL;   /* current opened file name in workspace */
+    ec->basename = NULL;   /* use this to form .dvi/.ps/.log etc. files */
     ec->pdffile = NULL;
     ec->workfile = NULL;
     ec->bibfile = NULL;
-    ec->tmpdir = g_strdup (g_get_tmp_dir ());
-
+    ec->projfile = NULL;
+    
     GtkSourceLanguageManager* manager = gtk_source_language_manager_new ();
     GtkSourceLanguage* lang = gtk_source_language_manager_get_language (manager,
             "latex");
     ec->buffer = gtk_source_buffer_new_with_language (lang);
-    ec->view =
-        GTK_SOURCE_VIEW (gtk_source_view_new_with_buffer (ec->buffer));
+    ec->view = GTK_SOURCE_VIEW (gtk_source_view_new_with_buffer (ec->buffer));
     ec->stylemanager = gtk_source_style_scheme_manager_get_default ();
     ec->errortag = gtk_text_tag_new ("error");
     ec->searchtag = gtk_text_tag_new ("search");
@@ -85,12 +91,9 @@ GuEditor* editor_init (GuMotion* mc) {
     gtk_source_view_set_tab_width (ec->view,
             atoi (config_get_value ("tabwidth")));
     gtk_source_view_set_insert_spaces_instead_of_tabs (ec->view,
-            (gboolean)config_get_value ("spaces_instof_tabs"));
+            TO_BOOL (config_get_value ("spaces_instof_tabs")));
     gtk_source_view_set_auto_indent (ec->view,
-            (gboolean)config_get_value ("autoindentation"));
-
-    const gchar* style_scheme = config_get_value ("style_scheme");
-    editor_set_style_scheme_by_id (ec, style_scheme);
+            TO_BOOL (config_get_value ("autoindentation")));
 
 #ifdef USE_GTKSPELL
     if (config_get_value ("spelling"))
@@ -105,6 +108,13 @@ GuEditor* editor_init (GuMotion* mc) {
                 G_CALLBACK (on_key_press_cb), mc);
     ec->sigid[1] = g_signal_connect (ec->view, "key-release-event",
                 G_CALLBACK (on_key_release_cb), mc);
+    ec->sigid[2] = g_signal_connect (ec->buffer, "changed",
+                G_CALLBACK (check_preview_timer), NULL);
+
+    ec->sigid[3] = g_signal_connect_after(ec->buffer, "insert-text", 
+                G_CALLBACK(on_inserted_text), ec);
+    ec->sigid[4] = g_signal_connect_after(ec->buffer, "delete-range", 
+                G_CALLBACK(on_delete_range), ec);
 
     return ec;
 }
@@ -114,20 +124,35 @@ void editor_destroy (GuEditor* ec) {
 
     for (i = 0; i < 2; ++i)
         g_signal_handler_disconnect (ec->view, ec->sigid[i]);
+    for (i = 2; i < 5; ++i)
+        g_signal_handler_disconnect (ec->buffer, ec->sigid[i]);
 
     editor_fileinfo_cleanup (ec);
+    g_free(ec);
 }
 
-gboolean editor_fileinfo_update_biblio (GuEditor* ec,  const gchar* filename) {
-    g_free (ec->bibfile);
+static void on_inserted_text(GtkTextBuffer *textbuffer,GtkTextIter *location,
+                             gchar *text,gint len, gpointer user_data) {
 
-    if (ec->filename && !g_path_is_absolute (filename)) {
-        gchar* dirname = g_path_get_dirname (ec->filename);
-        ec->bibfile = g_build_filename (dirname, filename, NULL);
-        g_free (dirname);
-    } else
-        ec->bibfile = g_strdup (filename);
-    return utils_path_exists (ec->bibfile);
+    if (location == NULL || user_data == NULL) {
+        return;
+    }
+    GuEditor* e = GU_EDITOR(user_data);
+    
+    e->last_edit = *location;
+    e->sync_to_last_edit = TRUE;
+}
+
+static void on_delete_range(GtkTextBuffer *textbuffer,GtkTextIter *start,
+                             GtkTextIter *end, gpointer user_data) {
+
+    if (start == NULL || user_data == NULL) {
+        return;
+    }
+    GuEditor* e = GU_EDITOR(user_data);
+    
+    e->last_edit = *start;
+    e->sync_to_last_edit = TRUE;
 }
 
 /* FileInfo:
@@ -158,35 +183,51 @@ void editor_fileinfo_update (GuEditor* ec, const gchar* filename) {
     if (ec->workfd != -1)
         editor_fileinfo_cleanup (ec);
 
-    ec->fdname = g_build_filename (ec->tmpdir, "gummi_XXXXXX", NULL);
-    ec->workfd = g_mkstemp (ec->fdname); 
+    ec->fdname = g_build_filename (C_TMPDIR, "gummi_XXXXXX", NULL);
+    ec->workfd = g_mkstemp (ec->fdname);
 
     if (filename) {
-        gchar* basename = g_path_get_basename (filename);
-        gchar* dirname = g_path_get_dirname (filename);
+        gchar* base = g_path_get_basename (filename);
+        gchar* dir = g_path_get_dirname (filename);
         ec->filename = g_strdup (filename);
-        ec->workfile = g_strdup_printf ("%s%c.%s.swp", dirname, G_DIR_SEPARATOR,
-                                       basename);
-        ec->pdffile =  g_strdup_printf ("%s%c.%s.pdf", ec->tmpdir,
-                                       G_DIR_SEPARATOR, basename);
-        g_free (basename);
-        g_free (dirname);
+        ec->basename = g_strdup_printf ("%s%c.%s", dir, G_DIR_SEPARATOR, base);
+        ec->workfile = g_strdup_printf ("%s.swp", ec->basename);
+        ec->pdffile =  g_strdup_printf ("%s%c.%s.pdf", C_TMPDIR,
+                                       G_DIR_SEPARATOR, base);
+        g_free (base);
+        g_free (dir);
     } else {
         ec->workfile = g_strdup (ec->fdname);
+        ec->basename = g_strdup (ec->fdname);
         ec->pdffile =  g_strdup_printf ("%s.pdf", ec->fdname);
     }
 }
 
+gboolean editor_fileinfo_update_biblio (GuEditor* ec,  const gchar* filename) {
+    g_free (ec->bibfile);
+
+    if (ec->filename && !g_path_is_absolute (filename)) {
+        gchar* dirname = g_path_get_dirname (ec->filename);
+        ec->bibfile = g_build_filename (dirname, filename, NULL);
+        g_free (dirname);
+    } else
+        ec->bibfile = g_strdup (filename);
+    return utils_path_exists (ec->bibfile);
+}
+
 void editor_fileinfo_cleanup (GuEditor* ec) {
     gchar* auxfile = NULL;
     gchar* logfile = NULL;
-
+    gchar* syncfile = NULL;
+    
     if (ec->filename) {
         gchar* dirname = g_path_get_dirname (ec->filename);
         gchar* basename = g_path_get_basename (ec->filename);
-        auxfile = g_strdup_printf ("%s%c.%s.aux", ec->tmpdir,
+        auxfile = g_strdup_printf ("%s%c.%s.aux", C_TMPDIR,
                 G_DIR_SEPARATOR, basename);
-        logfile = g_strdup_printf ("%s%c.%s.log", ec->tmpdir,
+        logfile = g_strdup_printf ("%s%c.%s.log", C_TMPDIR,
+                G_DIR_SEPARATOR, basename);
+        syncfile = g_strdup_printf ("%s%c.%s.synctex.gz", C_TMPDIR,
                 G_DIR_SEPARATOR, basename);
         g_free (basename);
         g_free (dirname);
@@ -195,30 +236,38 @@ void editor_fileinfo_cleanup (GuEditor* ec) {
         gchar* basename = g_path_get_basename (ec->workfile);
         auxfile = g_strdup_printf ("%s.aux", ec->fdname);
         logfile = g_strdup_printf ("%s.log", ec->fdname);
+        syncfile = g_strdup_printf ("%s.synctex.gz", ec->fdname);
         g_free (basename);
         g_free (dirname);
     }
 
+    // TODO: make a loop or maybe make register of created files? proc? 
+
     close (ec->workfd);
     ec->workfd = -1;
 
     g_remove (auxfile);
     g_remove (logfile);
+    g_remove (syncfile);
     g_remove (ec->fdname);
     g_remove (ec->workfile);
     g_remove (ec->pdffile);
+    g_remove (ec->basename);
 
     g_free (auxfile);
     g_free (logfile);
+    g_free (syncfile);
     g_free (ec->fdname);
     g_free (ec->filename);
     g_free (ec->workfile);
     g_free (ec->pdffile);
+    g_free (ec->basename);
 
     ec->fdname = NULL;
     ec->filename = NULL;
     ec->workfile = NULL;
     ec->pdffile = NULL;
+    ec->basename = NULL;
 }
 
 void editor_sourceview_config (GuEditor* ec) {
@@ -226,6 +275,9 @@ void editor_sourceview_config (GuEditor* ec) {
 
     gtk_source_buffer_set_highlight_matching_brackets (ec->buffer, TRUE);
 
+    const gchar* style_scheme = config_get_value ("style_scheme");
+    editor_set_style_scheme_by_id (ec, style_scheme);
+
     const gchar* font = config_get_value ("font");
     slog (L_INFO, "setting font to %s\n", font);
     PangoFontDescription* font_desc = pango_font_description_from_string (font);
@@ -233,9 +285,9 @@ void editor_sourceview_config (GuEditor* ec) {
     pango_font_description_free (font_desc);
 
     gtk_source_view_set_show_line_numbers (GTK_SOURCE_VIEW (ec->view),
-            (gboolean)config_get_value ("line_numbers"));
+            TO_BOOL (config_get_value ("line_numbers")));
     gtk_source_view_set_highlight_current_line (GTK_SOURCE_VIEW (ec->view),
-        (gboolean)config_get_value ("highlighting"));
+        TO_BOOL (config_get_value ("highlighting")));
 
     /* The condition 'textwrapping=False && wordwrapping=True' can't happen */
     if (config_get_value ("textwrapping"))
@@ -244,9 +296,6 @@ void editor_sourceview_config (GuEditor* ec) {
         wrapmode += 1;
 
     gtk_text_view_set_wrap_mode (ec_view, wrapmode);
-    g_object_set (G_OBJECT (ec->errortag), "background", "red",
-            "foreground", "white", NULL);
-    g_object_set (G_OBJECT (ec->searchtag), "background", "yellow", NULL);
 }
 
 
@@ -287,17 +336,23 @@ void editor_fill_buffer (GuEditor* ec, const gchar* text) {
     gtk_text_buffer_get_start_iter (ec_buffer, &start);
     gtk_text_buffer_place_cursor (ec_buffer, &start);
     gtk_widget_grab_focus (GTK_WIDGET (ec->view));
+    ec->sync_to_last_edit = FALSE;
 }
 
 gchar* editor_grab_buffer (GuEditor* ec) {
     GtkTextIter start, end;
-    gtk_widget_set_sensitive (GTK_WIDGET (ec->view), FALSE);
     gtk_text_buffer_get_bounds (ec_buffer, &start, &end);
     gchar* pstr = gtk_text_iter_get_text (&start, &end);
-    gtk_widget_set_sensitive (GTK_WIDGET (ec->view), TRUE);
     return pstr;
 }
 
+gboolean editor_buffer_changed (GuEditor* ec) {
+    if (gtk_text_buffer_get_modified (GTK_TEXT_BUFFER (ec->buffer))) {
+        return TRUE;
+    }
+    return FALSE;
+}
+
 void editor_insert_package (GuEditor* ec, const gchar* package) {
     GtkTextIter start, mstart, mend, sstart, send;
     gchar* pkgstr = g_strdup_printf ("\\usepackage{%s}\n", package);
@@ -308,8 +363,7 @@ void editor_insert_package (GuEditor* ec, const gchar* package) {
                 NULL)) {
         gtk_source_buffer_begin_not_undoable_action (ec->buffer);
         gtk_text_buffer_begin_user_action (ec_buffer);
-        gtk_text_buffer_insert (ec_buffer, &mstart,
-                pkgstr, strlen (pkgstr));
+        gtk_text_buffer_insert (ec_buffer, &mstart, pkgstr, -1);
         gtk_text_buffer_end_user_action (ec_buffer);
         gtk_source_buffer_end_not_undoable_action (ec->buffer);
         gtk_text_buffer_set_modified (ec_buffer, TRUE);
@@ -329,8 +383,7 @@ void editor_insert_bib (GuEditor* ec, const gchar* package) {
                 &sstart, &send, NULL)) {
         gtk_source_buffer_begin_not_undoable_action (ec->buffer);
         gtk_text_buffer_begin_user_action (ec_buffer);
-        gtk_text_buffer_insert (ec_buffer, &mstart,
-                pkgstr, strlen (pkgstr));
+        gtk_text_buffer_insert (ec_buffer, &mstart, pkgstr, -1);
         gtk_text_buffer_end_user_action (ec_buffer);
         gtk_source_buffer_end_not_undoable_action (ec->buffer);
         gtk_text_buffer_set_modified (ec_buffer, TRUE);
@@ -356,13 +409,13 @@ void editor_set_selection_textstyle (GuEditor* ec, const gchar* type) {
 
     /* select style */
     for (i = 0; i < style_size; ++i)
-        if (0 == strcmp (style[i][0], type)) {
+        if (0 == g_strcmp0 (style[i][0], type)) {
             selected = i;
             break;
         }
 
     /* generate regex expression */
-    regexbuf = g_strdup_printf (" (.*)%s%s (.*)%s%s (.*)",
+    regexbuf = g_strdup_printf ("(.*)%s%s(.*)%s%s(.*)",
             (style[selected][1][0] == '\\')? "\\": "",
             style[selected][1],
             (style[selected][2][0] == '\\')? "\\": "",
@@ -393,7 +446,7 @@ void editor_set_selection_textstyle (GuEditor* ec, const gchar* type) {
 
     gtk_text_buffer_begin_user_action (ec_buffer);
     gtk_text_buffer_delete (ec_buffer, &start, &end);
-    gtk_text_buffer_insert (ec_buffer, &start, outtext, strlen (outtext));
+    gtk_text_buffer_insert (ec_buffer, &start, outtext, -1);
     end = start;
     gtk_text_iter_backward_chars (&start, strlen (outtext));
     gtk_text_buffer_select_range (ec_buffer, &start, &end);
@@ -548,7 +601,7 @@ void editor_start_replace_next (GuEditor* ec, const gchar* term,
             && gtk_text_iter_ends_word (&mend)))) {
         gtk_text_buffer_begin_user_action (ec_buffer);
         gtk_text_buffer_delete (ec_buffer, &mstart, &mend);
-        gtk_text_buffer_insert (ec_buffer, &mstart, rterm, strlen (rterm));
+        gtk_text_buffer_insert (ec_buffer, &mstart, rterm, -1);
         gtk_text_buffer_end_user_action (ec_buffer);
         editor_search_next (ec, FALSE);
     }
@@ -573,8 +626,7 @@ void editor_start_replace_all (GuEditor* ec, const gchar* term,
                 && gtk_text_iter_ends_word (&mend)))) {
             gtk_text_buffer_begin_user_action (ec_buffer);
             gtk_text_buffer_delete (ec_buffer, &mstart, &mend);
-            gtk_text_buffer_insert (ec_buffer, &mstart, rterm,
-                    strlen (rterm));
+            gtk_text_buffer_insert (ec_buffer, &mstart, rterm, -1);
             gtk_text_buffer_end_user_action (ec_buffer);
             start =  mstart;
         } else break;
@@ -592,6 +644,17 @@ void editor_scroll_to_cursor (GuEditor* ec) {
                                  0.25, FALSE, 0.0, 0.0);
 }
 
+void editor_scroll_to_line (GuEditor* ec, gint line) {
+    if (ec == NULL) {
+        return;
+    }
+    GtkTextIter iter;
+    gtk_text_buffer_get_iter_at_line(ec_buffer, &iter, line);
+    gtk_text_buffer_place_cursor(ec_buffer, &iter);
+    editor_scroll_to_cursor(ec);
+    ec->sync_to_last_edit = FALSE;
+}
+
 void editor_undo_change (GuEditor* ec) {
     GtkTextIter current;
     if (gtk_source_buffer_can_undo (ec->buffer)) {
@@ -613,6 +676,7 @@ void editor_redo_change (GuEditor* ec) {
 }
 
 void editor_set_style_scheme_by_id (GuEditor* ec, const gchar* id) {
+    
     GtkSourceStyleScheme* scheme =
         gtk_source_style_scheme_manager_get_scheme (ec->stylemanager, id);
     slog (L_INFO, "setting styles scheme to %s\n", id);
@@ -623,6 +687,117 @@ void editor_set_style_scheme_by_id (GuEditor* ec, const gchar* id) {
                 "classic");
     }
     gtk_source_buffer_set_style_scheme (ec->buffer, scheme);
+    
+    set_style_fg_bg(G_OBJECT (ec->searchtag), scheme, "search-match", "yellow");
+    set_style_fg_bg(G_OBJECT (ec->errortag), scheme,"def:error",  "red");
+}
+
+
+static inline gdouble gdkcolor_luminance(GdkColor c) {
+    return (0.2126 * c.red + 0.7152 * c.green + 0.0722 * c.blue) / G_MAXUINT16;
+}
+
+
+/**
+ *  Sets a object's fore- and background color to that of scheme's style 
+ *  "styleName". If no background color is defined in the style, defaultBG is 
+ *  used. defaultBG can be any valid parameter to gdk_color_parse().
+ *  If only a foreground color was defined and it has not enough contrast to the
+ *  default background, it will be overwritten. The foreground color will either
+ *  be white or black, which has more contrast.
+ */
+void set_style_fg_bg (GObject* obj, GtkSourceStyleScheme* scheme, 
+                      gchar* styleName, gchar* defaultBG)
+{
+    GtkSourceStyle *style = NULL;
+    
+    gchar *bg = NULL;
+    gchar *fg = NULL;
+    gboolean foreground_set;
+    gboolean background_set;
+    GdkColor foreground;
+    GdkColor background;
+
+
+    if (scheme == NULL) {
+        goto set_style_fg_bg_return_defaults;
+    }
+    
+    style = gtk_source_style_scheme_get_style (scheme, styleName);
+    
+    if (style == NULL) {
+        goto set_style_fg_bg_return_defaults;
+    }
+    
+    // Get properties of style
+    g_object_get (style, 
+                  "foreground-set", &foreground_set, 
+                  "foreground", &fg,
+                  "background-set", &background_set, 
+                  "background", &bg,
+                  NULL);
+
+    if (foreground_set) {
+        if (fg == NULL || !gdk_color_parse (fg, &foreground))
+            foreground_set = FALSE;
+    }
+
+    if (background_set) {
+        if (bg == NULL || !gdk_color_parse (bg, &background))
+            background_set = FALSE;
+    }    
+
+    g_free(fg);
+    g_free(bg);
+    
+    if (background_set && foreground_set) {
+        // We trust the style to set both to good values
+        // Do nothing
+    } else if (!background_set && foreground_set) {
+        // Set bg to default and check if fg has enough contrast
+        gdk_color_parse(defaultBG, &background);
+        gdouble diff = ABS(gdkcolor_luminance(foreground) -
+                gdkcolor_luminance(background));
+        if (diff < 0.5) {
+            slog(L_INFO, "Style \"%s\" defines a foreground, but no background "
+                         "color. As the fourground color has not enough "
+                         "contrast to Gummis default background color, the "
+                         "foreground color has been adjusted.\n", styleName);
+            if (gdkcolor_luminance(background) > 0.5) {
+                gdk_color_parse("black", &foreground);
+            } else {
+                gdk_color_parse("white", &foreground);
+            }
+        }
+    } else if (background_set && !foreground_set) {
+        // Choose a fg = white or black, which has more contrast   
+        if (gdkcolor_luminance(background) > 0.5) {
+            gdk_color_parse("black", &foreground);
+        } else {
+            gdk_color_parse("white", &foreground);
+        }
+    } else {
+        // none set, set defaults
+        goto set_style_fg_bg_return_defaults;
+    }
+    
+    g_object_set (obj, "foreground-gdk", &foreground, 
+                        "background-gdk", &background, NULL);
+    return;
+                        
+set_style_fg_bg_return_defaults:
+
+    // No valid style, set defaults
+    gdk_color_parse(defaultBG, &background);
+    if (gdkcolor_luminance(background) > 0.5) {
+        gdk_color_parse("black", &foreground);
+    } else {
+        gdk_color_parse("white", &foreground);
+    }
+    
+    g_object_set (obj, "foreground-gdk", &foreground, 
+                        "background-gdk", &background, NULL);
+                        
 }
 
 /* The following functions are taken from gedit and partially modified */
@@ -637,16 +812,17 @@ gint schemes_compare (gconstpointer a, gconstpointer b) {
     return g_utf8_collate (name_a, name_b);
 }
 
-GList* editor_list_style_scheme_sorted (GuEditor* ec) {
+GList* editor_list_style_scheme_sorted (void) {
     const gchar * const * scheme_ids;
     GList *schemes = NULL;
+    GtkSourceStyleSchemeManager* manager =
+        gtk_source_style_scheme_manager_get_default();
 
-    scheme_ids = gtk_source_style_scheme_manager_get_scheme_ids (
-            ec->stylemanager);
+    scheme_ids = gtk_source_style_scheme_manager_get_scheme_ids (manager);
 
     while (*scheme_ids != NULL) {
         GtkSourceStyleScheme *scheme;
-        scheme = gtk_source_style_scheme_manager_get_scheme (ec->stylemanager,
+        scheme = gtk_source_style_scheme_manager_get_scheme (manager,
                 *scheme_ids);
         schemes = g_list_prepend (schemes, scheme);
         ++scheme_ids;
