@@ -2,7 +2,7 @@
  * @file    snippets.c
  * @brief   handle snppets
  *
- * Copyright (C) 2010 Gummi-Dev Team <alexvandermey@gmail.com>
+ * Copyright (C) 2010-2011 Gummi-Dev Team <alexvandermey@gmail.com>
  * All Rights reserved.
  * 
  * Permission is hereby granted, free of charge, to any person
@@ -52,6 +52,8 @@ GuSnippets* snippets_init (const gchar* filename) {
 
     s->filename = g_strdup (filename);
     s->accel_group = gtk_accel_group_new ();
+    s->stackframe = NULL;
+
     snippets_load (s);
     return s;
 }
@@ -99,12 +101,15 @@ void snippets_load (GuSnippets* sc) {
         return;
     }
 
-    current = sc->head = prev = g_new0 (slist, 1);
-
     while (fgets (buf, BUFSIZ, fh)) {
+        current = g_new0 (slist, 1);
+        if (!sc->head)
+            sc->head = prev = current;
+        prev->next = current;
         buf[strlen (buf) -1] = 0; /* remove trailing '\n' */
+
         if (buf[0] != '\t') {
-            if ('#' == buf[0]) {
+            if ('#' == buf[0] || !strlen(buf)) {
                 current->first = g_strdup (buf);
             } else {
                 seg = strstr (buf, " ") + 1;
@@ -124,11 +129,8 @@ void snippets_load (GuSnippets* sc) {
             current = prev;
         }
         prev = current;
-        current->next = g_new0 (slist, 1);
-        current = current->next;
     }
-    g_free (current);
-    prev->next = NULL;
+    if (prev) prev->next = NULL;
     fclose (fh);
 }
 
@@ -143,7 +145,7 @@ void snippets_save (GuSnippets* sc) {
 
     while (current) {
         /* skip comments */
-        if ('#' == current->first[0]) {
+        if (!current->second) {
             fputs (current->first, fh);
             fputs ("\n", fh);
             current = current->next;
@@ -225,6 +227,7 @@ void snippets_set_accelerator (GuSnippets* sc, gchar* config) {
 
 void snippets_activate (GuSnippets* sc, GuEditor* ec, gchar* key) {
     gchar* snippet = NULL;
+    GuSnippetInfo* new_info = NULL;
     GtkTextIter start, end;
 
     slog (L_DEBUG, "Snippet `%s' activated\n", key);
@@ -232,46 +235,72 @@ void snippets_activate (GuSnippets* sc, GuEditor* ec, gchar* key) {
     snippet = snippets_get_value (sc, key);
     g_return_if_fail (snippet != NULL);
 
-    editor_get_current_iter (ec, &start);
-    sc->info = snippets_parse (snippet);
-    sc->info->start_offset = gtk_text_iter_get_offset (&start);
+    new_info = snippets_parse (snippet);
 
     gtk_text_buffer_get_selection_bounds (ec_buffer, &start, &end);
-    sc->info->sel_text = gtk_text_iter_get_text (&start, &end);
+    new_info->start_offset = gtk_text_iter_get_offset (&start);
+    new_info->sel_text = gtk_text_iter_get_text (&start, &end);
     GSList* marks = gtk_text_iter_get_marks (&start);
-    sc->info->sel_start = *GTK_TEXT_MARK (marks->data);
+    new_info->sel_start = *GTK_TEXT_MARK (marks->data);
     g_slist_free (marks);
 
-    gtk_text_buffer_insert (ec_buffer, &start, sc->info->expanded, -1);
-    snippet_info_create_marks (sc->info, ec);
-    snippet_info_initial_expand (sc->info, ec);
+    gtk_text_buffer_insert (ec_buffer, &start, new_info->expanded, -1);
+    snippet_info_create_marks (new_info, ec);
+    snippet_info_initial_expand (new_info, ec);
     gtk_text_buffer_set_modified (ec_buffer, TRUE);
 
-    sc->activated = TRUE;
+    if (sc->info) {
+        snippet_info_sync_group (sc->info, ec);
+        sc->stackframe = g_list_append (sc->stackframe, sc->info);
+    }
+    sc->info = new_info;
+
     if (!snippet_info_goto_next_placeholder (sc->info, ec))
         snippets_deactivate (sc, ec);
 }
 
 void snippets_deactivate (GuSnippets* sc, GuEditor* ec) {
-    sc->activated = FALSE;
+    GList* last = NULL;
     snippet_info_free (sc->info, ec);
+    last = g_list_last (sc->stackframe);
+    if (last) {
+        sc->info = last->data;
+        sc->stackframe = g_list_remove (sc->stackframe, last->data);
+    } else
+        sc->info = NULL;
     slog (L_DEBUG, "Snippet deactivated\n");
 }
 
 gboolean snippets_key_press_cb (GuSnippets* sc, GuEditor* ec, GdkEventKey* ev) {
     GtkTextIter current, start;
 
-    if (ev->keyval != GDK_KEY_Tab && !sc->activated)
-       return FALSE;
+    if (ev->keyval == GDK_KEY_Tab) {
+        gchar* key = NULL;
+        editor_get_current_iter (ec, &current);
+        if (gtk_text_iter_ends_word (&current)) {
+            start = current;
+            gtk_text_iter_backward_word_start (&start);
+            key = gtk_text_iter_get_text (&start, &current);
 
-    if (sc->activated) {
+            if (snippets_get_value (sc, key)) {
+                gtk_text_buffer_delete (ec_buffer, &start, &current);
+                snippets_activate (sc, ec, key);
+                g_free (key);
+                return TRUE;
+            }
+            g_free (key);
+        }
+    }
+    
+    if (sc->info) {
         if (ev->keyval == GDK_KEY_Tab) {
             if (!snippet_info_goto_next_placeholder (sc->info, ec))
                 snippets_deactivate (sc, ec);
             return TRUE;
         } else if (ev->keyval == GDK_KEY_ISO_Left_Tab
-                && ev->state & GDK_SHIFT_MASK) {
-            snippet_info_goto_prev_placeholder (sc->info, ec);
+                   && ev->state & GDK_SHIFT_MASK) {
+            if (!snippet_info_goto_prev_placeholder (sc->info, ec))
+                snippets_deactivate (sc, ec);
             return TRUE;
         }
         /* Deactivate snippet if the current insert range is not within the
@@ -286,34 +315,19 @@ gboolean snippets_key_press_cb (GuSnippets* sc, GuEditor* ec, GdkEventKey* ev) {
             if (offset < sc->info->start_offset || offset > bound_end)
                 snippets_deactivate (sc, ec);
         }
-    } else {
-        gchar* key = NULL;
-        gboolean found = FALSE;
-        editor_get_current_iter (ec, &current);
-        if (!gtk_text_iter_ends_word (&current)) return FALSE;
-        start = current;
-        gtk_text_iter_backward_word_start (&start);
-        key = gtk_text_iter_get_text (&start, &current);
-
-        if (snippets_get_value (sc, key)) {
-            gtk_text_buffer_delete (ec_buffer, &start, &current);
-            snippets_activate (sc, ec, key);
-            found = TRUE;
-        }
-        g_free (key);
-        return found;
     }
+
     return FALSE;
 }
 
-gboolean snippets_key_release_cb (GuSnippets* sc, GuEditor* ec, GdkEventKey* e) {
-    if (e->keyval != GDK_KEY_Tab && !sc->activated)
-       return FALSE;
-
-    if (sc->activated) {
-        snippet_info_sync_group (sc->info, ec);
+gboolean snippets_key_release_cb (GuSnippets* sc, GuEditor* ec, GdkEventKey* e)
+{
+    if (e->keyval != GDK_KEY_Tab && !sc->info)
         return FALSE;
-    }
+
+    if (sc->info)
+        snippet_info_sync_group (sc->info, ec);
+
     return FALSE;
 }
 
@@ -323,11 +337,16 @@ GuSnippetInfo* snippets_parse (char* snippet) {
     gchar** result = NULL;
     GRegex* regex = NULL;
     GMatchInfo* match_info = NULL;
-    const gchar* holders[] = { "\\$([0-9]+)", "\\${([0-9]*):?([^}]*)}" };
+    const gchar* holders[] = { "\\$([0-9]+)", "\\${([0-9]*):?([^}]*)}",
+                               "\\$(FILENAME)", "\\${(FILENAME)}",
+                               "\\$(BASENAME)", "\\${(BASENAME)}",
+                               "\\$(SELECTED_TEXT)", "\\${(SELECTED_TEXT)}"
+                               /* Anyway to combine these? */
+                            };
 
     GuSnippetInfo* info = snippet_info_new (snippet);
 
-    for (i = 0; i < 2; ++i) {
+    for (i = 0; i < sizeof(holders) / sizeof(holders[0]); ++i) {
         if (! (regex = g_regex_new (holders[i], G_REGEX_DOTALL, 0, &err))) {
             slog (L_ERROR, "g_regex_new (): %s\n", err->message);
             g_error_free (err);
@@ -337,8 +356,13 @@ GuSnippetInfo* snippets_parse (char* snippet) {
         while (g_match_info_matches (match_info)) {
             result = g_match_info_fetch_all (match_info);
             g_match_info_fetch_pos (match_info, 0, &start, &end);
-            snippet_info_append_holder (info, atoi (result[1]), start,
-                    end -start, result[2]);
+            if (i < 2) {
+                snippet_info_append_holder (info, atoi (result[1]), start,
+                        end -start, result[2]);
+            } else {
+                snippet_info_append_holder (info, -1, start, end -start,
+                        result[1]);
+            }
             slog (L_DEBUG, "Placeholder: (%s, %s, %s)\n", result[0], result[1],
                     result[2]);
             g_match_info_next (match_info, NULL);
@@ -350,6 +374,7 @@ GuSnippetInfo* snippets_parse (char* snippet) {
     info->einfo = g_list_sort (info->einfo, snippet_info_pos_cmp);
     info->einfo_sorted = g_list_copy (info->einfo);
     info->einfo_sorted = g_list_sort (info->einfo_sorted, snippet_info_num_cmp);
+
     return info;
 }
 
@@ -361,8 +386,7 @@ void snippets_accel_cb (GtkAccelGroup* accel_group, GObject* obj,
      * gtk_accel_group must be connect when load, we can not specify the
      * editor in user_data, because snippets should only have effect on
      * active tab */
-    if (!sc->activated)
-        snippets_activate (sc, gummi_get_active_editor (), key);
+    snippets_activate (sc, gummi_get_active_editor (), key);
 }
 
 void snippets_accel_connect (GuSnippets* sc, guint keyval, GdkModifierType mod,
@@ -387,7 +411,7 @@ void snippets_accel_disconnect (GuSnippets* sc, const gchar* key) {
     current = sc->closure_data;
     while (current) {
         closure_data = TUPLE2 (current->data);
-        if (strcmp (closure_data->first, key) == 0)
+        if (g_strcmp0 (closure_data->first, key) == 0)
             break;
         current = g_list_next (current);
     }
@@ -429,24 +453,29 @@ gboolean snippet_info_goto_next_placeholder (GuSnippetInfo* info, GuEditor* ec)
 
     /* Snippet just activated */
     if (!info->current) {
-        /* Skip $0 and jump to next placeholder */
-        if (info->einfo_unique
-            && GU_SNIPPET_EXPAND_INFO (info->einfo_unique->data)->group_number
-            == 0)
-            info->current = g_list_next (info->einfo_unique);
-        else
+        /* Skip $0, $-1 and jump to next placeholder */
+        if (info->einfo_unique) {
             info->current = info->einfo_unique;
+            while (info->current &&
+                GU_SNIPPET_EXPAND_INFO (info->current->data)->group_number <= 0)
+                info->current = g_list_next (info->current);
+        }
     } else
         info->current = g_list_next (info->current);
 
     /* No placeholder left */
     if (!info->current) {
         info->current = g_list_first (info->einfo_sorted);
-        if (!info->current
-            || GU_SNIPPET_EXPAND_INFO (info->current->data)->group_number != 0)
+        while (info->current &&
+               GU_SNIPPET_EXPAND_INFO (info->current->data)->group_number != 0)
+            info->current = g_list_next(info->current);
+
+        if (!info->current)
             return FALSE;
-        success = FALSE;
+        else /* This is the last one($0) set to false to deactivate snippet */
+            success = FALSE;
     }
+
     einfo = GU_SNIPPET_EXPAND_INFO (info->current->data);
     gtk_text_buffer_get_iter_at_mark (ec_buffer, &start, einfo->left_mark);
     gtk_text_buffer_get_iter_at_mark (ec_buffer, &end, einfo->right_mark);
@@ -455,18 +484,23 @@ gboolean snippet_info_goto_next_placeholder (GuSnippetInfo* info, GuEditor* ec)
     return success;
 }
 
-void snippet_info_goto_prev_placeholder (GuSnippetInfo* info, GuEditor* ec) {
+gboolean snippet_info_goto_prev_placeholder (GuSnippetInfo* info, GuEditor* ec)
+{
     GuSnippetExpandInfo* einfo = NULL;
     GtkTextIter start, end;
-    if (!info->current)
-        return;
-    else
-        info->current = g_list_previous (info->current);
+    info->current = g_list_previous (info->current);
+
+    /* Return false to deactivate snippet */
+    if (!info->current ||
+        GU_SNIPPET_EXPAND_INFO(info->current->data)->group_number < 0)
+        return FALSE;
+
     einfo = GU_SNIPPET_EXPAND_INFO (info->current->data);
     gtk_text_buffer_get_iter_at_mark (ec_buffer, &start, einfo->left_mark);
     gtk_text_buffer_get_iter_at_mark (ec_buffer, &end, einfo->right_mark);
     gtk_text_buffer_place_cursor (ec_buffer, &start);
     gtk_text_buffer_select_range (ec_buffer, &start, &end);
+    return TRUE;
 }
 
 void snippet_info_append_holder (GuSnippetInfo* info, gint group, gint start,
@@ -515,6 +549,7 @@ void snippet_info_initial_expand (GuSnippetInfo* info, GuEditor* ec) {
     GtkTextIter start, end;
     GHashTable* map = NULL;
     GList* current = NULL;
+    gchar* text = NULL;
     gint key = 0;
 
     map = g_hash_table_new (NULL, NULL);
@@ -522,7 +557,7 @@ void snippet_info_initial_expand (GuSnippetInfo* info, GuEditor* ec) {
 
     while (current) {
         GuSnippetExpandInfo* einfo = GU_SNIPPET_EXPAND_INFO (current->data);
-        if (!g_hash_table_lookup_extended (map, ( (gpointer)einfo->group_number),
+        if (!g_hash_table_lookup_extended (map, ((gpointer)einfo->group_number),
                     (gpointer)&key, (gpointer)&value)) {
             g_hash_table_insert (map, (gpointer)einfo->group_number, einfo);
             info->einfo_unique = g_list_append (info->einfo_unique, einfo);
@@ -541,7 +576,9 @@ void snippet_info_initial_expand (GuSnippetInfo* info, GuEditor* ec) {
         gtk_text_buffer_get_iter_at_mark (ec_buffer, &start, einfo->left_mark);
         gtk_text_buffer_get_iter_at_mark (ec_buffer, &end, einfo->right_mark);
 
-        if (strcmp (value->text, "$SELECTED_TEXT") == 0) {
+        /* Expand macros */
+        text = GU_SNIPPET_EXPAND_INFO(current->data)->text;
+        if (g_strcmp0 (text, "SELECTED_TEXT") == 0) {
             GtkTextIter ms, me;
             gtk_text_buffer_delete (ec_buffer, &start, &end);
             gtk_text_buffer_insert (ec_buffer, &start, info->sel_text, -1);
@@ -549,26 +586,31 @@ void snippet_info_initial_expand (GuSnippetInfo* info, GuEditor* ec) {
             me = ms;
             gtk_text_iter_forward_chars (&me, strlen (info->sel_text));
             gtk_text_buffer_delete (ec_buffer, &ms, &me);
-        } else if (strcmp (value->text, "$FILENAME") == 0) {
+        } else if (g_strcmp0 (text, "FILENAME") == 0) {
             gtk_text_buffer_delete (ec_buffer, &start, &end);
             gtk_text_buffer_insert (ec_buffer, &start,
                     ec->filename? ec->filename: "", -1);
-        } else if (strcmp (value->text, "$BASENAME") == 0) {
-            gchar* basename = g_path_get_basename (ec->filename?ec->filename:"");
+        } else if (g_strcmp0 (text, "BASENAME") == 0) {
+            gchar* basename = g_path_get_basename(ec->filename?ec->filename:"");
             gtk_text_buffer_delete (ec_buffer, &start, &end);
             gtk_text_buffer_insert (ec_buffer, &start, basename, -1);
             g_free (basename);
         } else {
+            /* Expand text of same group with with text of group leader */
             gtk_text_buffer_delete (ec_buffer, &start, &end);
             gtk_text_buffer_insert (ec_buffer, &start, value->text, -1);
         }
+
         current = g_list_next (current);
     }
     g_hash_table_destroy (map);
 }
 
 void snippet_info_sync_group (GuSnippetInfo* info, GuEditor* ec) {
-    if (!info->current) return;
+    if (!info->current ||
+        GU_SNIPPET_EXPAND_INFO(info->current->data)->group_number == -1)
+        return;
+
     GuSnippetExpandInfo* active = GU_SNIPPET_EXPAND_INFO (info->current->data);
     GList* current = g_list_first (info->einfo);
     gchar* text = NULL;
